import ase

import plumed_nodes as pn

s1 = pn.IndexSelector(indices=[0])
s2 = pn.IndexSelector(indices=[1])
distance_cv = pn.DistanceCV(x1=s1, x2=s2, prefix="d12", group_reduction="com")
print_action = pn.PrintActionNode(cv=distance_cv, stride=10, file="output.txt")

bias = pn.MetaDBiasCV(cv=distance_cv, sigma=0.1, grid_min=0.0, grid_max=10.0)

meta_d = pn.MetaDynamicsConfig(height=0.5, pace=150)

meta_d_model = pn.MetaDynamicsModel(
    config=meta_d, bias_cvs=[bias], data=ase.Atoms(), actions=[print_action]
)



import dataclasses
import ase
from plumed_nodes.interfaces import CollectiveVariable


@dataclasses.dataclass
class PrintCVAction:
    """Node for PRINT action."""

    cvs: list[CollectiveVariable]
    stride: int = 1
    file: str = "COLVAR"

    def to_plumed(self, atoms: ase.Atoms) -> list[str]:
        """Convert the action node to a PLUMED input string."""
        commands = []
        for cv in self.cvs:
            labels, commands = cv.to_plumed(atoms)
            if not commands:
                raise ValueError(f"Empty PLUMED commands for CV {cv.prefix}")

            print_command = f"PRINT ARG={','.join(labels)} STRIDE={self.stride} FILE={self.file}"
            commands.append(print_command)
        return commands



import pytest
import rdkit2ase

import plumed_nodes


@pytest.fixture(scope="session")
def ethanol_water():
    ethanol = rdkit2ase.smiles2conformers("CCO", numConfs=100)
    water = rdkit2ase.smiles2conformers("O", numConfs=100)
    box = rdkit2ase.pack(
        [ethanol, water], counts=[16, 16], density=700, packmol="packmol.jl"
    )
    return box.copy()


@pytest.fixture(scope="session")
def small_ethnol_water():
    ethanol = rdkit2ase.smiles2conformers("CCO", numConfs=1)
    water = rdkit2ase.smiles2conformers("O", numConfs=1)
    box = rdkit2ase.pack(
        [ethanol, water], counts=[2, 2], density=700, packmol="packmol.jl"
    )
    return box.copy()



from ase.calculators.calculator import Calculator, all_changes
from ase.calculators.plumed import Plumed



class NonOverwritingPlumed(Plumed):
    def calculate(self, atoms=None, properties=None, system_changes=all_changes):
        if properties is None:
            properties = ["energy", "forces"]
        Calculator.calculate(self, atoms, properties, system_changes)
        energy, forces = self.compute_energy_and_forces(
            self.atoms.get_positions(), self.istep
        )
        self.istep += 1
        self.results = {f"model_{k}": v for k, v in self.calc.results.items()}
        self.results["energy"], self.results["forces"] = energy, forces


import plumed_nodes as pn


def test_index_selector(ethanol_water):
    indices = [0, 1, 2]
    selector = pn.IndexSelector(indices=indices)
    assert selector.select(ethanol_water) == [[0], [1], [2]]


def test_smiles_selector(small_ethnol_water):
    # ethanol
    selector = pn.SMILESSelector(smiles="CCO")
    selected_indices = selector.select(small_ethnol_water)
    assert selected_indices == [list(range(9)), list(range(9, 18))]
    # water
    selector = pn.SMILESSelector(smiles="O")
    selected_indices = selector.select(small_ethnol_water)
    assert selected_indices == [[18, 19, 20], [21, 22, 23]]



import pathlib
from typing import Protocol

import ase
from ase.calculators.calculator import Calculator
from PIL import Image


class NodeWithCalculator(Protocol):
    """Any class with a `get_calculator` method returning an ASE Calculator."""

    def get_calculator(
        self, *, directory: str | pathlib.Path | None = None, **kwargs
    ) -> Calculator: ...


class AtomSelector(Protocol):
    """Protocol for selecting atoms within a single ASE Atoms object.

    This interface defines the contract for selecting atoms based on various
    criteria within an individual frame/structure.
    """

    def select(self, atoms: ase.Atoms) -> list[list[int]]:
        """Select atoms based on the implemented criteria.

        Parameters
        ----------
        atoms : ase.Atoms
            The atomic structure to select from.

        Returns
        -------
        list[list[int]]
            Groups of indices of selected atoms. All indices in the inner lists
            are representative of the same group, e.g. one molecule.
        """
        ...

class PlumedGenerator(Protocol):
    """Protocol for generating PLUMED strings from collective variables."""

    def to_plumed(self, atoms: ase.Atoms) -> list[str]: ...


class CollectiveVariable(Protocol):
    """Protocol for collective variables (CVs) that can be used in PLUMED."""

    prefix: str

    def get_img(self, atoms: ase.Atoms) -> Image.Image: ...

    def to_plumed(self, atoms: ase.Atoms) -> tuple[list[str], list[str]]:
        """
        Convert the collective variable to a PLUMED string.

        Parameters
        ----------
        atoms : ase.Atoms
            The atomic structure to use for generating the PLUMED string.

        Returns
        -------
        tuple[list[str], str]
            - List of distance labels.
            - list of PLUMED strings representing the CV.
        """
        ...


class MetadynamicsBiasCollectiveVariable(Protocol):
    """Protocol for metadata associated with a bias in PLUMED."""

    cv: CollectiveVariable
    sigma: float | None = None
    grid_min: float | None = None
    grid_max: float | None = None
    grid_bin: int | None = None



import dataclasses
from typing import Union, Literal

from ase import Atoms
from rdkit import Chem
from rdkit.Chem import Draw
from PIL import Image
import rdkit2ase

from plumed_nodes.interfaces import AtomSelector, CollectiveVariable

# Define allowed string literals for type hints
GroupReductionStrategyType = Literal["com", "cog", "first", "all"]
MultiGroupStrategyType = Literal["first", "all_pairs", "corresponding", "first_to_all"]


@dataclasses.dataclass
class DistanceCV(CollectiveVariable):
    """
    PLUMED DISTANCE collective variable.
    """

    x1: AtomSelector
    x2: AtomSelector
    prefix: str
    group_reduction: GroupReductionStrategyType = "com"
    multi_group: MultiGroupStrategyType = "first"
    create_virtual_sites: bool = True

    def get_img(self, atoms: Atoms) -> Image.Image | None:
        """
        Generates an image of the molecule(s) with the selected atoms for the distance CV highlighted.

        - The first atom from the first selector (`x1`) is colored red.
        - The first atom from the second selector (`x2`) is colored blue.

        If the atoms are in different molecular fragments, both fragments are shown side-by-side.

        Returns
        -------
        PIL.Image.Image | None
            A PIL Image object of the visualization, or None if atoms could not be selected.
        """
        # --- Start of new/modified code ---
        
        # 1. Get atom groups from each selector
        groups1 = self.x1.select(atoms)
        groups2 = self.x2.select(atoms)

        if not groups1 or not groups1[0] or not groups2 or not groups2[0]:
            print("Warning: One or both selectors returned no atoms for visualization.")
            return None

        # For visualization, we only highlight the first atom of the first group from each selector
        atom1_idx = groups1[0][0]
        atom2_idx = groups2[0][0]

        # 2. Convert ASE object to RDKit and identify molecular fragments
        mol = rdkit2ase.ase2rdkit(atoms)
        mol_frags = Chem.GetMolFrags(mol, asMols=True)
        frag_indices_list = Chem.GetMolFrags(mol, asMols=False)

        # 3. Find which fragments contain the selected atoms and prepare for drawing
        mols_to_draw = []
        highlights_to_draw = []
        colors_to_draw = []
        red = (1.0, 0.2, 0.2)
        blue = (0.2, 0.2, 1.0)

        for i, frag_mol in enumerate(mol_frags):
            # Create a map of global (original) index -> local (fragment) index
            local_idx_map = {global_idx: local_idx for local_idx, global_idx in enumerate(frag_indices_list[i])}

            current_highlights = {}
            
            # Check if our selected atoms are in this fragment
            if atom1_idx in local_idx_map:
                current_highlights[local_idx_map[atom1_idx]] = red
            if atom2_idx in local_idx_map:
                current_highlights[local_idx_map[atom2_idx]] = blue

            # If this fragment contains at least one of the selected atoms, add it for drawing
            if current_highlights:
                mols_to_draw.append(frag_mol)
                highlights_to_draw.append(list(current_highlights.keys()))
                colors_to_draw.append(current_highlights)
        
        if not mols_to_draw:
            return None

        # 4. Generate the image using RDKit's grid drawing utility
        img = Draw.MolsToGridImage(
            mols_to_draw,
            molsPerRow=len(mols_to_draw),  # Places all fragments in a single row
            subImgSize=(400, 400),
            highlightAtomLists=highlights_to_draw,
            highlightAtomColors=colors_to_draw,
            useSVG=False  # Ensures a PIL Image object is returned
        )
        return img

    def to_plumed(self, atoms: Atoms) -> tuple[list[str], list[str]]:
        """Generate PLUMED input string(s) for DISTANCE.

        Returns
        -------
        - list of distance labels
        - PLUMED input string
        """
        groups1 = self.x1.select(atoms)
        groups2 = self.x2.select(atoms)

        if not groups1 or not groups2:
            raise ValueError(f"Empty selection for distance CV {self.prefix}")

        # Check for overlaps
        overlaps = self._check_overlaps(groups1, groups2)
        if overlaps and self.group_reduction not in ["com", "cog"]:
            raise ValueError(
                f"Overlapping atoms found: {overlaps}. "
                "This is only valid with CENTER_OF_MASS or CENTER_OF_GEOMETRY reduction."
            )

        commands = self._generate_commands(groups1, groups2, atoms)

        # Extract labels from commands
        labels = []
        for cmd in commands:
            if ":" in cmd and cmd.strip().startswith((self.prefix, f"{self.prefix}_")):
                label_part = cmd.split(":")[0].strip()
                if "DISTANCE" in cmd:
                    labels.append(label_part)

        return labels, commands

    def _check_overlaps(
        self, groups1: list[list[int]], groups2: list[list[int]]
    ) -> set[int]:
        """Check for overlapping indices between groups."""
        flat1 = {idx for group in groups1 for idx in group}
        flat2 = {idx for group in groups2 for idx in group}
        return flat1.intersection(flat2)

    def _generate_commands(
        self, groups1: list[list[int]], groups2: list[list[int]], atoms: Atoms
    ) -> list[str]:
        """Generate PLUMED commands based on the strategies."""
        commands = []

        # Determine which groups to process based on multi_group strategy
        if self.multi_group == "first":
            # Only process first groups
            process_groups1 = [groups1[0]]
            process_groups2 = [groups2[0]]
            group_pairs = [(0, 0)]
        elif self.multi_group == "all_pairs":
            process_groups1 = groups1
            process_groups2 = groups2
            group_pairs = [
                (i, j) for i in range(len(groups1)) for j in range(len(groups2))
            ]
        elif self.multi_group == "corresponding":
            n = min(len(groups1), len(groups2))
            process_groups1 = groups1[:n]
            process_groups2 = groups2[:n]
            group_pairs = [(i, i) for i in range(n)]
        elif self.multi_group == "first_to_all":
            process_groups1 = [groups1[0]]
            process_groups2 = groups2
            group_pairs = [(0, j) for j in range(len(groups2))]

        # Create virtual sites for the groups we're actually using
        sites1 = {}
        sites2 = {}

        for i, group in enumerate(process_groups1):
            site, site_commands = self._reduce_group(
                group, f"{self.prefix}_g1_{i}", atoms
            )
            sites1[i] = site
            commands.extend(site_commands)

        for j, group in enumerate(process_groups2):
            site, site_commands = self._reduce_group(
                group, f"{self.prefix}_g2_{j}", atoms
            )
            sites2[j] = site
            commands.extend(site_commands)

        # Create distance commands for the specified pairs
        for i, j in group_pairs:
            if len(group_pairs) == 1:
                dist_label = self.prefix
            else:
                dist_label = f"{self.prefix}_{i}_{j}"

            commands.append(
                self._make_distance_command(sites1[i], sites2[j], dist_label)
            )

        return commands

    def _reduce_group(
        self, group: list[int], prefix: str, atoms: Atoms
    ) -> tuple[Union[str, list[int]], list[str]]:
        """
        Reduce a group to a single point based on reduction strategy.

        Returns:
            - Site identifier (atom index, virtual site label, or group)
            - List of PLUMED commands to create virtual sites
        """
        commands = []

        if len(group) == 1:
            # Single atom - no reduction needed
            return str(group[0] + 1), commands

        if self.group_reduction == "first":
            return str(group[0] + 1), commands

        if self.group_reduction == "com":
            if self.create_virtual_sites:
                site_label = f"{prefix}_com"
                atom_list = ",".join(str(idx + 1) for idx in group)
                commands.append(f"{site_label}: COM ATOMS={atom_list}")
                return site_label, commands
            else:
                return group, commands

        if self.group_reduction == "cog":
            site_label = f"{prefix}_cog"
            atom_list = ",".join(str(idx + 1) for idx in group)
            commands.append(f"{site_label}: CENTER ATOMS={atom_list}")
            return site_label, commands

        if self.group_reduction == "all":
            return group, commands

        raise ValueError(f"Unknown group reduction strategy: {self.group_reduction}")

    def _make_distance_command(
        self, site1: Union[str, list], site2: Union[str, list], label: str
    ) -> str:
        """Create a single DISTANCE command."""
        if isinstance(site1, str) and isinstance(site2, str):
            return f"{label}: DISTANCE ATOMS={site1},{site2}"
        elif isinstance(site1, list) and isinstance(site2, list):
            atoms1 = ",".join(str(idx + 1) for idx in site1)
            atoms2 = ",".join(str(idx + 1) for idx in site2)
            return f"{label}: DISTANCE ATOMS1={atoms1} ATOMS2={atoms2}"
        else:
            if isinstance(site1, list):
                atoms1 = ",".join(str(idx + 1) for idx in site1)
                return f"{label}: DISTANCE ATOMS1={atoms1} ATOMS2={site2}"
            else:
                atoms2 = ",".join(str(idx + 1) for idx in site2)
                return f"{label}: DISTANCE ATOMS1={site1} ATOMS2={atoms2}"



from plumed_nodes.actions import PrintCVAction
from plumed_nodes.cvs import DistanceCV
from plumed_nodes.metadynamics import MetaDBiasCV, MetaDynamicsConfig, MetaDynamicsModel
from plumed_nodes.selectors import IndexSelector, SMILESSelector, SMARTSSelector

__all__ = [
    "PrintCVAction",
    "DistanceCV",
    "IndexSelector",
    "SMILESSelector",
    "SMARTSSelector",
    "MetaDynamicsModel",
    "MetaDBiasCV",
    "MetaDynamicsConfig",
]



import ase

import plumed_nodes as pn


def test_distance_cv_first_strategy(small_ethnol_water):
    """Test default FIRST strategy - only first groups."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=x1_selector,
        x2=x2_selector,
        prefix="d12",
        multi_group="first"
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethnol_water)

    expected = [
        "d12_g1_0_com: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_g2_0_com: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_g1_0_com,d12_g2_0_com"
    ]

    assert plumed_str == expected
    assert labels == ["d12"]


def test_distance_cv_all_pairs(small_ethnol_water):
    """Test ALL_PAIRS strategy - all combinations."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=x1_selector,
        x2=x2_selector,
        prefix="d",
        multi_group="all_pairs"
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethnol_water)

    # Should create 2x2=4 distances
    expected = [
        "d_g1_0_com: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d_g1_1_com: COM ATOMS=10,11,12,13,14,15,16,17,18",
        "d_g2_0_com: COM ATOMS=19,20,21",
        "d_g2_1_com: COM ATOMS=22,23,24",
        "d_0_0: DISTANCE ATOMS=d_g1_0_com,d_g2_0_com",
        "d_0_1: DISTANCE ATOMS=d_g1_0_com,d_g2_1_com",
        "d_1_0: DISTANCE ATOMS=d_g1_1_com,d_g2_0_com",
        "d_1_1: DISTANCE ATOMS=d_g1_1_com,d_g2_1_com"
    ]

    assert plumed_str == expected
    assert labels == ["d_0_0", "d_0_1", "d_1_0", "d_1_1"]


def test_distance_cv_corresponding(small_ethnol_water):
    """Test CORRESPONDING strategy - pair by index."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=x1_selector,
        x2=x2_selector,
        prefix="d",
        multi_group="corresponding"
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethnol_water)

    expected = [
        "d_g1_0_com: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d_g1_1_com: COM ATOMS=10,11,12,13,14,15,16,17,18",
        "d_g2_0_com: COM ATOMS=19,20,21",
        "d_g2_1_com: COM ATOMS=22,23,24",
        "d_0_0: DISTANCE ATOMS=d_g1_0_com,d_g2_0_com",
        "d_1_1: DISTANCE ATOMS=d_g1_1_com,d_g2_1_com"
    ]

    assert plumed_str == expected
    assert labels == ["d_0_0", "d_1_1"]


def test_distance_cv_first_to_all(small_ethnol_water):
    """Test FIRST_TO_ALL strategy - first of x1 to all of x2."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=x1_selector,
        x2=x2_selector,
        prefix="d",
        multi_group="first_to_all"
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethnol_water)

    expected = [
        "d_g1_0_com: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d_g2_0_com: COM ATOMS=19,20,21",
        "d_g2_1_com: COM ATOMS=22,23,24",
        "d_0_0: DISTANCE ATOMS=d_g1_0_com,d_g2_0_com",
        "d_0_1: DISTANCE ATOMS=d_g1_0_com,d_g2_1_com"
    ]

    assert plumed_str == expected
    assert labels == ["d_0_0", "d_0_1"]


def test_distance_cv_single_atoms():
    """Test with single atom selections."""
    atoms = ase.Atoms("H2", positions=[[0, 0, 0], [1, 0, 0]])

    x1_selector = pn.IndexSelector(indices=[0])
    x2_selector = pn.IndexSelector(indices=[1])

    distance_cv = pn.DistanceCV(x1=x1_selector, x2=x2_selector, prefix="h_h")

    labels, plumed_str = distance_cv.to_plumed(atoms)

    expected = ["h_h: DISTANCE ATOMS=1,2"]
    assert plumed_str == expected
    assert labels == ["h_h"]


def test_distance_cv_first_atom_reduction(small_ethnol_water):
    """Test FIRST_ATOM reduction strategy."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=x1_selector,
        x2=x2_selector,
        prefix="d",
        group_reduction="first",
        multi_group="first"
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethnol_water)

    # Should use first atom of each group
    expected = ["d: DISTANCE ATOMS=1,19"]
    assert plumed_str == expected
    assert labels == ["d"]


def test_distance_cv_cog_reduction(small_ethnol_water):
    """Test CENTER_OF_GEOMETRY reduction."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=x1_selector,
        x2=x2_selector,
        prefix="d",
        group_reduction="cog",
        multi_group="first"
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethnol_water)

    # Should use CENTER instead of COM
    expected = ["d_g1_0_cog: CENTER ATOMS=1,2,3,4,5,6,7,8,9",
                "d_g2_0_cog: CENTER ATOMS=19,20,21",
                "d: DISTANCE ATOMS=d_g1_0_cog,d_g2_0_cog"]

    assert plumed_str == expected
    assert labels == ["d"]


def test_distance_cv_no_virtual_sites(small_ethnol_water):
    """Test with create_virtual_sites=False."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=x1_selector,
        x2=x2_selector,
        prefix="d",
        create_virtual_sites=False,
        multi_group="first",
    )

    labels, plumed_str = distance_cv.to_plumed(small_ethnol_water)

    expected = ["d: DISTANCE ATOMS1=1,2,3,4,5,6,7,8,9 ATOMS2=19,20,21"]
    assert plumed_str == expected
    assert labels == ["d"]




import dataclasses
from pathlib import Path

import ase.units
import zntrack
from ase.calculators.calculator import Calculator
from plumed_nodes.calc import NonOverwritingPlumed

from plumed_nodes.interfaces import (
    CollectiveVariable,
    MetadynamicsBiasCollectiveVariable,
    NodeWithCalculator,
    PlumedGenerator
)


@dataclasses.dataclass
class MetaDBiasCV(MetadynamicsBiasCollectiveVariable):
    """
    Resources
    ---------
    - https://www.plumed.org/doc-master/user-doc/html/METAD/
    """
    cv: CollectiveVariable
    sigma: float | None = None
    grid_min: float | None = None
    grid_max: float | None = None
    grid_bin: int | None = None


@dataclasses.dataclass
class MetaDynamicsConfig:
    """
    Base configuration for metadynamics.
    This contains only the global parameters that apply to all CVs.
    """

    height: float = 1.0  # kJ/mol
    pace: int = 500
    biasfactor: float | None = None
    temp: float = 300.0
    file: str = "HILLS"
    adaptive: str = "NONE"  # NONE, DIFF, GEOM


class MetaDynamicsModel(zntrack.Node, NodeWithCalculator):
    config: MetaDynamicsConfig = zntrack.deps()
    data: list[ase.Atoms] = zntrack.deps()
    data_idx: int = zntrack.params(-1)
    bias_cvs: list[MetaDBiasCV] = zntrack.deps(default_factory=list)
    actions: list[PlumedGenerator] = zntrack.deps(default_factory=list)
    timestep: float = zntrack.params(1.0)  # in fs, default is 1 fs
    model: NodeWithCalculator = zntrack.deps()

    figures: Path = zntrack.outs_path(zntrack.nwd / "figures")

    def run(self):
        self.figures.mkdir(parents=True, exist_ok=True)
        for cv in self.bias_cvs:
            img = cv.cv.get_img(self.data[self.data_idx])
            img.save(self.figures / f"{cv.cv.prefix}.png")

    def get_calculator(
        self, *, directory: str | Path | None = None, **kwargs
    ) -> NonOverwritingPlumed:
        if directory is None:
            raise ValueError("Directory must be specified for PLUMED input files.")
        directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)

        lines = self.to_plumed(self.data[self.data_idx])
        # replace FILE= with f"FILE={directory}/" inside config
        lines = [line.replace("FILE=", f"FILE={directory}/") for line in lines]

        # Write plumed input file
        with (directory / "plumed.dat").open("w") as file:
            for line in lines:
                file.write(line + "\n")

        kT = ase.units.kB * self.config.temp

        return NonOverwritingPlumed(
            calc=self.model.get_calculator(directory=directory),
            atoms=self.data[self.data_idx],
            input=lines,
            timestep=float(self.timestep * ase.units.fs),
            kT=float(kT),
            log=(directory / "plumed.log").as_posix(),
        )

    def to_plumed(self, atoms: ase.Atoms) -> list[str]:
        """Generate PLUMED input string for the metadynamics model."""
        plumed_lines = []
        all_labels = []

        sigmas, grid_mins, grid_maxs, grid_bins = [], [], [], []

        plumed_lines.append(f"UNITS LENGTH=A TIME={1 / (1000 * ase.units.fs)} ENERGY={ase.units.mol / ase.units.kJ}")

        for bias_cv in self.bias_cvs:
            labels, cv_str = bias_cv.cv.to_plumed(atoms)
            plumed_lines.extend(cv_str)
            all_labels.extend(labels)

            # Collect per-CV parameters for later
            sigmas.append(str(bias_cv.sigma) if bias_cv.sigma is not None else None)
            grid_mins.append(str(bias_cv.grid_min) if bias_cv.grid_min is not None else None)
            grid_maxs.append(str(bias_cv.grid_max) if bias_cv.grid_max is not None else None)
            grid_bins.append(str(bias_cv.grid_bin) if bias_cv.grid_bin is not None else None)

        metad_parts = [
            "METAD",
            f"ARG={','.join(all_labels)}",
            f"HEIGHT={self.config.height}",
            f"PACE={self.config.pace}",
            f"TEMP={self.config.temp}",
            f"FILE={self.config.file}",
            f"ADAPTIVE={self.config.adaptive}",
        ]
        if self.config.biasfactor is not None:
            metad_parts.append(f"BIASFACTOR={self.config.biasfactor}")

        # Add SIGMA, GRID_MIN, GRID_MAX, GRID_BIN only if any value is set
        if any(v is not None for v in sigmas):
            metad_parts.append(f"SIGMA={','.join(v if v is not None else '0.0' for v in sigmas)}")
        if any(v is not None for v in grid_mins):
            metad_parts.append(f"GRID_MIN={','.join(v if v is not None else '0.0' for v in grid_mins)}")
        if any(v is not None for v in grid_maxs):
            metad_parts.append(f"GRID_MAX={','.join(v if v is not None else '0.0' for v in grid_maxs)}")
        if any(v is not None for v in grid_bins):
            metad_parts.append(f"GRID_BIN={','.join(v if v is not None else '0' for v in grid_bins)}")

        plumed_lines.append(f"metad: {' '.join(metad_parts)}")
        # Add actions
        for action in self.actions:
            action_lines = action.to_plumed(atoms)
            if not action_lines:
                raise ValueError(f"Empty PLUMED commands for action {action}")
            plumed_lines.extend(action_lines)
        return plumed_lines


import dataclasses

import ase
import rdkit2ase
import rdkit.Chem.rdchem as rdchem
import typing as tp
from rdkit import Chem

from plumed_nodes.interfaces import AtomSelector


@dataclasses.dataclass
class IndexSelector(AtomSelector):
    indices: list[int]

    def select(self, atoms: ase.Atoms) -> list[list[int]]:
        return [[x] for x in self.indices]


@dataclasses.dataclass
class SMILESSelector(AtomSelector):
    smiles: str

    def select(self, atoms: ase.Atoms) -> list[list[int]]:
        matches = rdkit2ase.match_substructure(atoms, smiles=self.smiles)
        return [list(match) for match in matches]

@dataclasses.dataclass
class SMARTSSelector(AtomSelector):
    """Select atoms based on SMARTS or mapped SMILES patterns.
    
    This selector uses RDKit's substructure matching to identify atoms
    matching a given SMARTS pattern or mapped SMILES string. It supports
    flexible hydrogen handling and can work with mapped atoms for
    precise selection.
    
    Parameters
    ----------
    pattern : str
        SMARTS pattern (e.g., "[F]", "[OH]", "C(=O)O") or SMILES with 
        atom maps (e.g., "C1[C:1]OC(=[O:1])O1"). If atom maps are present,
        only the mapped atoms are selected.
    hydrogens : {'exclude', 'include', 'isolated'}, default='exclude'
        How to handle hydrogen atoms in the selection:
        - 'exclude': Remove all hydrogens from the selection
        - 'include': Include hydrogens bonded to selected heavy atoms
        - 'isolated': Select only hydrogens bonded to selected heavy atoms
    
    Examples
    --------
    >>> # Select all fluorine atoms
    >>> selector = SMARTSSelection(pattern="[F]")
    
    >>> # Select carboxylic acid groups including hydrogens
    >>> selector = SMARTSSelection(pattern="C(=O)O", hydrogens="include")
    
    >>> # Select only specific mapped atoms
    >>> selector = SMARTSSelection(pattern="C1[C:1]OC(=[O:1])O1")
    """
    
    pattern: str
    hydrogens: tp.Literal["include", "exclude", "isolated"] = "exclude"
    
    def select(self, atoms: ase.Atoms) -> list[list[int]]:
        return rdkit2ase.select_atoms_grouped(rdkit2ase.ase2rdkit(atoms), self.pattern, self.hydrogens)



import plumed_nodes as pn


def test_distance_cv_corresponding_strategy(small_ethnol_water):
    """Test default corresponding strategy - only first groups."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=x1_selector,
        x2=x2_selector,
        prefix="d12",
        multi_group="corresponding"
    )

    biased_distance_cv = pn.MetaDBiasCV(
        cv=distance_cv,
        sigma=0.1,
        grid_min=0.0,
        grid_max=10.0,
        grid_bin=100,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=0.5,
        pace=150,
        biasfactor=10.0,
        temp=300.0,
        file="HILLS",
        adaptive="NONE",
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethnol_water,
        bias_cvs=[biased_distance_cv],
        actions=[pn.PrintCVAction(cv=distance_cv)],
        model=None # type: ignore
    )

    assert meta_d_model.to_plumed(small_ethnol_water) == [
        "UNITS LENGTH=A TIME=0.010180505671156723 ENERGY=96.48533288249877",
"d12_g1_0_com: COM ATOMS=1,2,3,4,5,6,7,8,9",
"d12_g1_1_com: COM ATOMS=10,11,12,13,14,15,16,17,18",
"d12_g2_0_com: COM ATOMS=19,20,21",
"d12_g2_1_com: COM ATOMS=22,23,24",
"d12_0_0: DISTANCE ATOMS=d12_g1_0_com,d12_g2_0_com",
"d12_1_1: DISTANCE ATOMS=d12_g1_1_com,d12_g2_1_com",
"metad: METAD ARG=d12_0_0,d12_1_1 HEIGHT=0.5 PACE=150 TEMP=300.0 FILE=HILLS ADAPTIVE=NONE BIASFACTOR=10.0 SIGMA=0.1 GRID_MIN=0.0 GRID_MAX=10.0 GRID_BIN=100",
'PRINT ARG=d12_0_0,d12_1_1 STRIDE=1 FILE=COLVAR'
    ]


def test_distance_cv_first_strategy(small_ethnol_water):
    """Test default FIRST strategy - only first groups."""
    x1_selector = pn.SMILESSelector(smiles="CCO")
    x2_selector = pn.SMILESSelector(smiles="O")

    distance_cv = pn.DistanceCV(
        x1=x1_selector,
        x2=x2_selector,
        prefix="d12",
        multi_group="first"
    )

    biased_distance_cv = pn.MetaDBiasCV(
        cv=distance_cv,
        sigma=0.1,
        grid_min=0.0,
        grid_max=10.0,
        grid_bin=100,
    )

    meta_d_config = pn.MetaDynamicsConfig(
        height=0.5,
        pace=150,
        biasfactor=10.0,
        temp=300.0,
        file="HILLS",
        adaptive="NONE",
    )

    meta_d_model = pn.MetaDynamicsModel(
        config=meta_d_config,
        data=small_ethnol_water,
        bias_cvs=[biased_distance_cv],
        model=None # type: ignore
    )

    assert meta_d_model.to_plumed(small_ethnol_water) == [
        "UNITS LENGTH=A TIME=0.010180505671156723 ENERGY=96.48533288249877",
        "d12_g1_0_com: COM ATOMS=1,2,3,4,5,6,7,8,9",
        "d12_g2_0_com: COM ATOMS=19,20,21",
        "d12: DISTANCE ATOMS=d12_g1_0_com,d12_g2_0_com",
        "metad: METAD ARG=d12 HEIGHT=0.5 PACE=150 TEMP=300.0 FILE=HILLS ADAPTIVE=NONE BIASFACTOR=10.0 SIGMA=0.1 GRID_MIN=0.0 GRID_MAX=10.0 GRID_BIN=100"
    ]
